(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.
Require Import Classical.

(* The Soundness and Completeness of Kripke semantics cannot be proved without classical logic *)

Require Import classical_help.
Require Import list_fall_exst.
Require Import replace.
  
Require Import predicate.
Require Import closure.
Require Import carrier.
Require Import maps.
Require Import finite.

Require Import formula.
Require Import labels.
Require Import words.
Require Import constraints.
Require Import css.
Require Import pmo.

Require Import kripke.

Require Import branch.
Require Import tableaux.

Require Import bi_expansion.
Require Import bi_exp_props.
Require Import bi_closed_branch.
Require Import bi_tableaux.

Section Oracles_def.

  Variable P : CSS_type -> Prop.
  
  Definition oracle_dwn_closed  := forall css1 css2, CSS css1 -> CSS_leq css1 css2 -> P css2 -> P css1.
  Definition oracle_finite_char := forall css, CSS css -> (forall cssf, finite_CSS cssf -> CSS_leq cssf css -> P cssf) -> P css.
  Definition oracle_open        := forall css, CSS css -> P css -> bi_closed_branch css -> False.
  Definition oracle_saturated   := forall css lcss, CSS css -> bi_expansion css lcss -> P css -> exists css', (carrier lcss) css' /\ P css'.

  Inductive oracle : Prop := 
    in_oracle : P inc CSS -> oracle_dwn_closed -> oracle_finite_char -> oracle_open -> oracle_saturated -> oracle.

End Oracles_def.

Definition CSS_consistent := bi_has_no_closed_tableau.
Definition CSS_fconsistent css := CSS css /\ forall cssf, finite_CSS cssf -> CSS_leq cssf css -> CSS_consistent cssf.

Section finite_sub_CSS.

  Ltac solve x := apply finite_dec_inc; [ apply x | intros; apply classic ].

  Let stm_inc_finite_lemma (X Y : stm -> Prop) : X inc Y -> finite Y -> finite X.         Proof. solve stm_eq_dec. Qed.
  Let cst_inc_finite_lemma (C D : cst -> Prop) : C inc D -> finite D -> finite C.         Proof. solve cst_eq_dec. Qed.

  (* with use classical reasonning to prove that a sub-CSS of a finite CSS is finite *)

  Proposition finite_sub_CSS : forall css1 css2, CSS css1 -> CSS_leq css1 css2 -> finite_CSS css2 -> finite_CSS css1.
  Proof.
     intros (X1,C1) (X2,C2) H0 [ H1 H2 ] [ H3 [ H4 H5] ].
     split; auto. split; auto.
     apply stm_inc_finite_lemma with (1 := H1); auto.
     apply cst_inc_finite_lemma with (1 := H2); auto.
  Qed.

End finite_sub_CSS.

Proposition CSS_consistent_closed X C X' C' : X inc X' -> C inc pmo C' -> finite_CSS (X,C) -> CSS_consistent (X',C') -> CSS_consistent (X,C).
Proof.
  intros H0 H1 H3 [ H4 H5 ].
  split; auto.  
  intros H; apply H5.
  apply bi_has_closed_tableau_pmo_monotonic with (5 := H); auto.
Qed.
      
Proposition CSS_consistent_down_closed : oracle_dwn_closed CSS_consistent.
Proof.
  intros (X,C) (X',C') H0 H1 [ H2 H3 ].
  assert (finite_CSS (X,C)) as H4.
    apply finite_sub_CSS with (3 := H2); auto.
  destruct H1 as [ H5 H6 ]; auto.
  apply CSS_consistent_closed with (X' := X') (C' := C'); auto.
  split; auto.
Qed.

Proposition finite_consistency_prop css : finite_CSS css -> (CSS_consistent css <-> CSS_fconsistent css).
Proof.
  intros [ H1 H2 ]; split; intros H3.

  split; auto.
  intros cssf [ H4 H5 ] H6.
  apply CSS_consistent_down_closed with (2 := H6); auto.
  
  destruct H3 as [ H3 H4 ].
  split. split; auto.
  apply (H4 css).
  split; auto.
  apply CSS_leq_refl; auto.
Qed.

Proposition fconsistent_CSS : CSS_fconsistent inc CSS.
Proof. intros ? []; auto. Qed.

Proposition not_fconsistent_CSS css : CSS css -> ~ CSS_fconsistent css -> exists cssf, finite_CSS cssf /\ CSS_leq cssf css /\ ~ CSS_consistent cssf.
Proof.
  intros H1 H2.
  unfold CSS_fconsistent in H2.
  classic_solve.
  destruct H2 as [ H2 | H2 ]; try contradiction (H2 H1).
  classic_solve.
  destruct H2 as [ cssf ]; exists cssf; auto.
Qed.

Proposition fCSS_not_consistent css : finite_CSS css -> ~ CSS_consistent css -> exists t, bi_tab css t /\ carrier t inc bi_closed_branch.
Proof.
  intros H1 H2.
  unfold CSS_consistent, bi_has_no_closed_tableau in H2.
  classic_solve.
  destruct H2 as [ H2 | H2 ]; try (contradict H2; auto; fail).
  classic_solve; auto.
Qed.

Proposition fconsistent_dwn_closed : oracle_dwn_closed CSS_fconsistent.
Proof.
  intros css1 css2 H1 H2 [ H3 H4 ].
  split; auto.
  intros cssf H6 H7.
  apply (H4 cssf); auto.
  apply CSS_leq_trans with (1 := H7); auto.
Qed.

Proposition fconsistent_finite_char : oracle_finite_char CSS_fconsistent.
Proof.
  intros css H1 H2.
  split; auto.
  intros cssf H4 H5.
  apply finite_consistency_prop; auto.
Qed.
 
Proposition fconsistent_open : oracle_open CSS_fconsistent.
Proof.
  intros css H1 [ H2 H3 ] H4.
  destruct (bi_cb_finite _ H4) as ( cssf & H5 & [ H7 H8 ] ).
  apply bin_inter_sg in H5; destruct H5 as [ H5 H6 ].
  destruct (H3 _ H8) as [ H9 H10 ]; auto.
  apply H10.
  exists (cssf::nil).
  split. 
  constructor 1; auto.
  apply carrier_cons_2; auto.
Qed.

Section saturated.

  (* There is some classical reasoning here as well *)

  Hint Resolve CSS_cup_rl CSS_cup_rr CSS_cup_l CSS_sg_l CSS_inc_empty_l CSS_dec_l without_sg_inc_l  without_pair_inc_l.

  Ltac split_disj :=
    match goal with 
        [ |- exists _, (?a = _ \/ False) /\ ?P _ ] => exists a; split; auto
      | [ |- exists _, (?a = _ \/ ?b = _ \/ False) /\ ?P _ ] => cut (P a \/ P b); [ intros []; [ exists a | exists b ]; auto | idtac ]
    end.

  Let nil_nil_lemma X (t : list X) : t = nil ++ t ++ nil.
  Proof. apply app_nil_end. Qed.

  Let cup_empty_lemma X (A B : X -> Prop) : A inc B cup empty -> A inc B.
  Proof. intros H; apply inc_transitive with (1 := H); auto. Qed.

  Ltac cup_tac := repeat
    match goal with 
      [ H: _ inc _ cup empty |- _ ] => apply cup_empty_lemma in H
    end.
  
  Proposition fconsistent_staturated : oracle_saturated CSS_fconsistent.
  Proof.
    generalize stm_eq_dec cst_eq_dec; intros Hstm Hcst.
    intros (X,C) lcss H1 [ r p Hr Hp ] [ _ H2 ].
    unfold bi_exp_rules in Hr; In_auto; symmetry in Hr; 
    subst; destruct p as ((((((A,B),a),b),m),x),y);
    simpl in Hp; try red in Hp; simpl; unfold expand_1;
    try (generalize (H1 _ _ _ Hp); intros Hp');
    split_disj.

    (* X (T A && B : m) *)
    
    split; auto.
    intros (Xf,Cf) [ H3 [ HXf XCf ] ] [ H4 H4' ].
    simpl in H4, H4'; cup_tac.
    destruct (pmo_cpct Hp') as (C0 & H5 & H6 & H7).
    set (Xf' := without_pair Xf (T A # m) (T B # m) cup sg (T A && B # m)).
    set (Cf' := Cf cup C0).

    assert (Xf' (T A && B # m)) as G0. unfold Xf'; pred_solve.
    assert (finite Xf') as G2. unfold Xf'; auto.
    assert (finite Cf') as G3. unfold Cf'; auto.
    assert (pmo Cf' (m~~m)) as G4. unfold Cf'; apply pmo_mono with (2 := H7); auto.
    assert (CSS (Xf',Cf')) as G1.
      unfold Xf', Cf'.
      apply CSS_cup_l; auto. 
      apply CSS_cup_rl, CSS_dec_l with (2 := H3); auto.
    assert (CSS_consistent (Xf',Cf')) as G5. unfold  Xf', Cf'; apply H2; auto.
    assert (CSS_consistent (Xf' cup (sg (T A # m) cup sg (T B # m)), Cf')) as G6.
      apply CSS_consistent_closed with (X' := Xf' cup (sg (T A # m) cup sg (T B # m))) (C' := Cf' cup empty); auto. 
      split; auto.
      intros (t & Ht1 & Ht2).
      generalize (bi_tableau_exp_Tand _ Cf' _ _ _ G0); intros G6. 
      destruct G5 as [ _ G5 ]; apply G5.
      exists t. split; auto.
      fold bi_tab; replace_with (bi_tab_comp _ nil nil _ _ Ht1 G6). 
    
    apply CSS_consistent_down_closed with (3 := G6); auto.
    split; auto; simpl.
    unfold Xf'; apply inc_transitive with (1 := without_pair_cup _ Hstm _ (T A # m) (T B # m)); auto.
    unfold Cf'; auto.

    (* X (F A && B : m) *)

    classic_disj H3 H4.
    apply not_fconsistent_CSS in H3; auto.
    apply not_fconsistent_CSS in H4; auto.
    destruct H4 as ( (XfB,CfB) & [ H61 [H62 H63] ] & [H71 H72] & H8 ).
    destruct H3 as ( (XfA,CfA) & [ H31 [H32 H33] ] & [H41 H42] & H5 ).
    destruct (pmo_cpct Hp') as (C0 & H9 & H10 & H11).
    simpl in H41, H42, H71, H72; cup_tac.
    set (Xf' := without_sg XfA (F A # m) cup without_sg XfB (F B # m) cup sg (F A && B # m)).
    set (Cf' := CfA cup CfB cup C0).

    assert (Xf' (F A && B # m)) as G0. unfold Xf'; pred_solve.
    assert (finite Xf') as G2. unfold Xf'; auto.
    assert (finite Cf') as G3. unfold Cf'; auto.
    assert (pmo Cf' (m~~m)) as G4. unfold Cf'; apply pmo_mono with (2 := H11); auto.
    assert (CSS (Xf',Cf')) as G1.
      unfold Xf', Cf'.
      apply CSS_cup_l; auto.
      apply CSS_cup_rl.
      apply CSS_cup_l; auto.
      apply CSS_cup_rl, CSS_dec_l with (2 := H31); auto.
      apply CSS_cup_rr, CSS_dec_l with (2 := H61); auto.
    assert (CSS_consistent (Xf',Cf')) as G5. unfold  Xf', Cf'; apply H2; auto.

    assert(~ CSS_consistent (Xf' cup sg (F A # m), Cf' cup empty)) as G6.
      unfold Xf', Cf'.
      intros H; apply H5.
      apply CSS_consistent_closed with (4 := H); auto.
      apply inc_transitive with (1 := without_sg_cup _ Hstm _ (F A # m)); pred_solve.
      apply inc_transitive with (2 := pmo_inc _); auto; pred_solve.
    
    assert(~ CSS_consistent (Xf' cup sg (F B # m), Cf' cup empty)) as G7.
      unfold Xf', Cf'.
      intros H; apply H8.
      apply CSS_consistent_closed with (4 := H); auto.
      apply inc_transitive with (1 := without_sg_cup _ Hstm _ (F B # m)); pred_solve.
      apply inc_transitive with (2 := pmo_inc _); auto; pred_solve.
  
    destruct fCSS_not_consistent with (2 := G6) as ( tA & HtA1 & HtA2 ); auto.
    destruct fCSS_not_consistent with (2 := G7) as ( tB & HtB1 & HtB2 ); auto.
 
    generalize (bi_tableau_exp_Fand _ Cf' _ _ _ G0); intros G8.
    generalize (bi_tab_comp _ nil _ _ _ HtA1 G8); simpl; intros G9.
    generalize (bi_tab_comp _ _ nil _ _ HtB1 G9); simpl; intros G10.
    rewrite <- app_nil_end in G10.
    
    assert (~ CSS_consistent (Xf', Cf')) as G11.
      intros H; apply H.
      exists (tA++tB).
      split; auto.
    apply G11.
    apply H2; auto.
    unfold Xf', Cf'; split; simpl; auto.

    (* X (T A %% B : m) *)

    classic_disj H3 H4.
    apply not_fconsistent_CSS in H3; auto.
    apply not_fconsistent_CSS in H4; auto.
    destruct H4 as ( (XfB,CfB) & [ H61 [H62 H63] ] & [H71 H72] & H8 ).
    destruct H3 as ( (XfA,CfA) & [ H31 [H32 H33] ] & [H41 H42] & H5 ).
    destruct (pmo_cpct Hp') as (C0 & H9 & H10 & H11).
    simpl in H41, H42, H71, H72; cup_tac.
    set (Xf' := without_sg XfA (T A # m) cup without_sg XfB (T B # m) cup sg (T A %% B # m)).
    set (Cf' := CfA cup CfB cup C0).

    assert (Xf' (T A %% B # m)) as G0. unfold Xf'; pred_solve.
    assert (finite Xf') as G2. unfold Xf'; auto.
    assert (finite Cf') as G3. unfold Cf'; auto.
    assert (pmo Cf' (m~~m)) as G4. unfold Cf'; apply pmo_mono with (2 := H11); auto.
    assert (CSS (Xf',Cf')) as G1.
      unfold Xf', Cf'.
      apply CSS_cup_l; auto.
      apply CSS_cup_rl.
      apply CSS_cup_l; auto.
      apply CSS_cup_rl, CSS_dec_l with (2 := H31); auto.
      apply CSS_cup_rr, CSS_dec_l with (2 := H61); auto.
    assert (CSS_consistent (Xf',Cf')) as G5. unfold  Xf', Cf'; apply H2; auto.

    assert(~ CSS_consistent (Xf' cup sg (T A # m), Cf' cup empty)) as G6.
      unfold Xf', Cf'.
      intros H; apply H5.
      apply CSS_consistent_closed with (4 := H); auto.
      apply inc_transitive with (1 := without_sg_cup _ Hstm _ (T A # m)); pred_solve.
      apply inc_transitive with (2 := pmo_inc _); auto; pred_solve.
    
    assert(~ CSS_consistent (Xf' cup sg (T B # m), Cf' cup empty)) as G7.
      unfold Xf', Cf'.
      intros H; apply H8.
      apply CSS_consistent_closed with (4 := H); auto.
      apply inc_transitive with (1 := without_sg_cup _ Hstm _ (T B # m)); pred_solve.
      apply inc_transitive with (2 := pmo_inc _); auto; pred_solve.
  
    destruct fCSS_not_consistent with (2 := G6) as ( tA & HtA1 & HtA2 ); auto.
    destruct fCSS_not_consistent with (2 := G7) as ( tB & HtB1 & HtB2 ); auto.
 
    generalize (bi_tableau_exp_Tor _ Cf' _ _ _ G0); intros G8.
    generalize (bi_tab_comp _ nil _ _ _ HtA1 G8); simpl; intros G9.
    generalize (bi_tab_comp _ _ nil _ _ HtB1 G9); simpl; intros G10.
    rewrite <- app_nil_end in G10.
    
    assert (~ CSS_consistent (Xf', Cf')) as G11.
      intros H; apply H.
      exists (tA++tB).
      split; auto.
    apply G11.
    apply H2; auto.
    unfold Xf', Cf'; split; simpl; auto.
    
    (* X (F A %% B : m) *)
    
    split; auto.
    intros (Xf,Cf) [ H3 [ HXf XCf ] ] [ H4 H4' ].
    simpl in H4, H4'; cup_tac.
    destruct (pmo_cpct Hp') as (C0 & H5 & H6 & H7).
    set (Xf' := without_pair Xf (F A # m) (F B # m) cup sg (F A %% B # m)).
    set (Cf' := Cf cup C0).

    assert (Xf' (F A %% B # m)) as G0. unfold Xf'; pred_solve.
    assert (finite Xf') as G2. unfold Xf'; auto.
    assert (finite Cf') as G3. unfold Cf'; auto.
    assert (pmo Cf' (m~~m)) as G4. unfold Cf'; apply pmo_mono with (2 := H7); auto.
    assert (CSS (Xf',Cf')) as G1.
      unfold Xf', Cf'.
      apply CSS_cup_l; auto. 
      apply CSS_cup_rl, CSS_dec_l with (2 := H3); auto.
    assert (CSS_consistent (Xf',Cf')) as G5. unfold  Xf', Cf'; apply H2; auto.
    assert (CSS_consistent (Xf' cup (sg (F A # m) cup sg (F B # m)), Cf')) as G6.
      apply CSS_consistent_closed with (X' := Xf' cup (sg (F A # m) cup sg (F B # m))) (C' := Cf' cup empty); auto. 
      split; auto.
      intros (t & Ht1 & Ht2).
      generalize (bi_tableau_exp_For _ Cf' _ _ _ G0); intros G6. 
      destruct G5 as [ _ G5 ]; apply G5.
      exists t. split; auto.
      fold bi_tab; replace_with (bi_tab_comp _ nil nil _ _ Ht1 G6). 
    
    apply CSS_consistent_down_closed with (3 := G6); auto.
    split; auto; simpl.
    unfold Xf'; apply inc_transitive with (1 := without_pair_cup _ Hstm _ (F A # m) (F B # m)); auto.
    unfold Cf'; auto.
    
    (* X (T A #> B : m) *)

    destruct Hp as [ Hp1 Hp2 ].
    classic_disj H3 H4.
    generalize (pmo_lft Hp2) (pmo_rt Hp2); intros Hp3 Hp4.
    apply not_fconsistent_CSS in H3; auto.
    apply not_fconsistent_CSS in H4; auto.
    destruct H4 as ( (XfB,CfB) & [ H61 [H62 H63] ] & [H71 H72] & H8 ).
    destruct H3 as ( (XfA,CfA) & [ H31 [H32 H33] ] & [H41 H42] & H5 ).
    destruct (pmo_cpct Hp2) as (C0 & H9 & H10 & H11).
    simpl in H41, H42, H71, H72; cup_tac.
    set (Xf' := without_sg XfA (F A # x) cup without_sg XfB (T B # x) cup sg (T A #> B # m)).
    set (Cf' := CfA cup CfB cup C0).

    assert (Xf' (T A #> B # m)) as G0. unfold Xf'; pred_solve.
    assert (finite Xf') as G2. unfold Xf'; auto.
    assert (finite Cf') as G3. unfold Cf'; auto.
    assert (pmo Cf' (m~~x)) as G4. unfold Cf'; apply pmo_mono with (2 := H11); auto.
    generalize (pmo_lft G4) (pmo_rt G4); intros G4' G4''.
    assert (CSS (Xf',Cf')) as G1.
      unfold Xf', Cf'.
      apply CSS_cup_l; auto.
      apply CSS_cup_rl.
      apply CSS_cup_l; auto.
      apply CSS_cup_rl, CSS_dec_l with (2 := H31); auto.
      apply CSS_cup_rr, CSS_dec_l with (2 := H61); auto.
    assert (CSS_consistent (Xf',Cf')) as G5. unfold  Xf', Cf'; apply H2; auto.

    assert(~ CSS_consistent (Xf' cup sg (F A # x), Cf' cup empty)) as G6.
      unfold Xf', Cf'.
      intros H; apply H5.
      apply CSS_consistent_closed with (4 := H); auto.
      apply inc_transitive with (1 := without_sg_cup _ Hstm _ (F A # x)); pred_solve.
      apply inc_transitive with (2 := pmo_inc _); auto; pred_solve.
    
    assert(~ CSS_consistent (Xf' cup sg (T B # x), Cf' cup empty)) as G7.
      unfold Xf', Cf'.
      intros H; apply H8.
      apply CSS_consistent_closed with (4 := H); auto.
      apply inc_transitive with (1 := without_sg_cup _ Hstm _ (T B # x)); pred_solve.
      apply inc_transitive with (2 := pmo_inc _); auto; pred_solve.
  
    destruct fCSS_not_consistent with (2 := G6) as ( tA & HtA1 & HtA2 ); auto.
    destruct fCSS_not_consistent with (2 := G7) as ( tB & HtB1 & HtB2 ); auto.
 
    generalize (bi_tableau_exp_Timp _ Cf' _ _ _ _ G0 G4); intros G8.
    generalize (bi_tab_comp _ nil _ _ _ HtA1 G8); simpl; intros G9.
    generalize (bi_tab_comp _ _ nil _ _ HtB1 G9); simpl; intros G10.
    rewrite <- app_nil_end in G10.
    
    assert (~ CSS_consistent (Xf', Cf')) as G11.
      intros H; apply H.
      exists (tA++tB).
      split; auto.
    apply G11.
    apply H2; auto.
    unfold Xf', Cf'; split; simpl; auto.
    
    (* X (F A #> B : m) *)
    
    destruct Hp as ( Hp1 & Hp2 ).
    generalize (H1 _ _ _ Hp1); intros Hp'.
    assert (pmo (C cup sg (m ~~ l b)) (l b ~~ l b)) as Hb.
      apply pmo_rt with (x := m), pmo_inc; pred_solve.
    
    split; auto.
    intros (Xf,Cf) [ H3 [ HXf XCf ] ] [ H4 H4' ].
    simpl in H4, H4'; cup_tac.
    destruct (pmo_cpct Hp') as (C0 & H5 & H6 & H7).
    set (Xf' := without_pair Xf (T A # l b) (F B # l b) cup sg (F A #> B # m)).
    assert (CSS (Xf',C)) as H8.
      unfold Xf'.
      apply CSS_dec_l with (2 := H1).
      apply without_pair_inc_l in H4.
      pred_solve.
    destruct CSS_compact with (1 := H8) as ( C1 & H9 & ( H10 & H11 & H12 ) ).
      unfold Xf'; auto.
    set (Cf' := without_sg Cf (m ~~ l b) cup (C1 cup C0)).

    assert (Xf' (F A #> B # m)) as G0. unfold Xf'; pred_solve.
    assert (finite Xf') as G2. unfold Xf'; auto.
    assert (finite Cf') as G3. unfold Cf'; auto.
    assert (CSS (Xf',Cf')) as G1. unfold Cf'; auto.
    assert (pmo Cf' (m~~m)) as G4. unfold Cf'; apply pmo_mono with (2 := H7); pred_solve.    
    assert (CSS_consistent (Xf',Cf')) as G5. unfold  Xf', Cf'; apply H2; auto.
    assert (~ alphabet Cf' b) as Hb'. unfold Cf'; intros H; apply Hp2, alphabet_inc with (2 := H); auto.
    assert (CSS_consistent (Xf' cup (sg (T A # l b) cup sg (F B # l b)), Cf' cup sg (m ~~ l b))) as G6.
      split; auto. split; auto.
      apply CSS_cup_l; auto.
      apply CSS_cup_l; auto; apply CSS_sg_l;
      apply pmo_rt with (x := m), pmo_inc; pred_solve.
      intros (t & Ht1 & Ht2). 
      generalize (bi_tableau_exp_Fimp _ Cf' _ _ _ _ G0 Hb'); intros G6. 
      generalize (bi_tab_comp _ nil nil _ _ Ht1 G6). 
      simpl; rewrite <- app_nil_end; intros G7.
      destruct G5 as [ _ G5 ]; apply G5. 
      exists t. split; auto.
    
    apply CSS_consistent_down_closed with (3 := G6); auto.
    split; simpl; auto.
    unfold Xf'; apply inc_transitive with (1 := without_pair_cup _ Hstm _ (T A # l b) (F B # l b)); auto.
    unfold Cf'; apply inc_transitive with (1 := without_sg_cup _ Hcst _ (m ~~ l b)); pred_solve.

    (* X (T I : m) *)

    split; auto.
    intros (Xf,Cf) [ H3 [ HXf XCf ] ] [ H4 H4' ].
    simpl in H4, H4'; cup_tac.
    destruct (pmo_cpct Hp') as (C0 & H5 & H6 & H7).
    assert (CSS (Xf, C)) as H8. intros s f n Hn; apply H4, H1 in Hn; auto.
    destruct (CSS_compact _ _ H8 HXf) as ( C1 & H9 & ( H10 & H11 & H12 ) ).
    set (Xf' := Xf cup sg (T I # m)).
    set (Cf' := without_sg Cf (eps~~m) cup (C1 cup C0)).

    assert (Xf' (T I # m)) as G0. unfold Xf'; pred_solve.
    assert (finite Xf') as G2. unfold Xf'; auto.
    assert (finite Cf') as G3. unfold Cf'; auto.
    assert (pmo Cf' (m~~m)) as G4. unfold Cf'; apply pmo_mono with (2 := H7); pred_solve.
    assert (CSS (Xf',Cf')) as G1. unfold Xf', Cf'; auto.
    assert (CSS_consistent (Xf',Cf')) as G5. unfold  Xf', Cf'; apply H2; auto.
 
    generalize (bi_tableau_exp_TI _ Cf' _ G0); intros G6.
    assert (CSS_consistent (Xf', Cf' cup sg (eps ~~ m))) as G7.
      apply CSS_consistent_closed with (X' := Xf' cup empty) (C' := Cf' cup sg (eps ~~m)); auto.
      repeat (split; auto).
      auto.
      intros ( t & Ht1 & Ht2 ).
      generalize (bi_tab_comp _ nil _ _ _ Ht1 G6); simpl; intros G7; rewrite <- app_nil_end in G7.      
      destruct G5 as [ _ G5 ]; apply G5.
      exists t; tauto.
  
    apply CSS_consistent_down_closed with (3 := G7); auto.
    unfold Xf', Cf'; repeat (split; simpl; auto).
    apply inc_transitive with (1 := without_sg_cup _ Hcst _ (eps ~~ m)); pred_solve.

    (* X (T A * B : m) *)
    
    destruct Hp as ( Hp1 & Hp2 & Hp3 & Hp4 ).
    generalize (H1 _ _ _ Hp1); intros Hp'.
    cutrewrite (a::b::nil = l a o l b); auto.
    assert (pmo (C cup sg (l a o l b ~~ m)) (l a ~~ l a)) as Ha.
      apply pmo_help_lemma_2 with (y := l b) (z := m), pmo_inc; pred_solve. 
    assert (pmo (C cup sg (l a o l b ~~ m)) (l b ~~ l b)) as Hb.
      apply pmo_help_lemma_1 with (x := l a) (z := m), pmo_inc; pred_solve.

    split; auto. 
    intros (Xf,Cf) [ H3 [ HXf XCf ] ] [ H4 H4' ].
    simpl in H4, H4'.
    destruct (pmo_cpct Hp') as (C0 & H5 & H6 & H7).
    set (Xf' := without_pair Xf (T A # l a) (T B # l b) cup sg (T A * B # m)).
    assert (CSS (Xf',C)) as H8.
      unfold Xf'.
      apply CSS_dec_l with (2 := H1).
      apply without_pair_inc_l in H4.
      pred_solve.
    destruct CSS_compact with (1 := H8) as ( C1 & H9 & ( H10 & H11 & H12 ) ).
      unfold Xf'; auto.
    set (Cf' := without_sg Cf (l a o l b ~~ m) cup (C1 cup C0)).

    assert (Xf' (T A * B # m)) as G0. unfold Xf'; pred_solve.
    assert (finite Xf') as G2. unfold Xf'; auto.
    assert (finite Cf') as G3. unfold Cf'; auto.
    assert (CSS (Xf',Cf')) as G1. unfold Cf'; auto.
    assert (pmo Cf' (m~~m)) as G4. unfold Cf'; apply pmo_mono with (2 := H7); pred_solve.    
    assert (CSS_consistent (Xf',Cf')) as G5. unfold  Xf', Cf'; apply H2; auto.
    assert (~ alphabet Cf' a) as Ha'. unfold Cf'; intros H; apply Hp3, alphabet_inc with (2 := H); auto.
    assert (~ alphabet Cf' b) as Hb'. unfold Cf'; intros H; apply Hp4, alphabet_inc with (2 := H); auto.
    assert (CSS_consistent (Xf' cup (sg (T A # l a) cup sg (T B # l b)), Cf' cup sg (l a o l b ~~ m))) as G6.
      split; auto. split; auto.
      apply CSS_cup_l; auto.
      apply CSS_cup_l; auto; apply CSS_sg_l.
      apply pmo_help_lemma_2 with (y := l b) (z := m), pmo_inc; pred_solve.
      apply pmo_help_lemma_1 with (x := l a) (z := m), pmo_inc; pred_solve.
      intros (t & Ht1 & Ht2). 
      generalize (bi_tableau_exp_Tmult _ Cf' _ _ _ _ _ G0 Hp2 Ha' Hb'); intros G6. 
      generalize (bi_tab_comp _ nil nil _ _ Ht1 G6). 
      simpl; rewrite <- app_nil_end; intros G7.
      destruct G5 as [ _ G5 ]; apply G5. 
      exists t. split; auto.
    
    apply CSS_consistent_down_closed with (3 := G6); auto.
    split; simpl; auto.
    unfold Xf'; apply inc_transitive with (1 := without_pair_cup _ Hstm _ (T A # l a) (T B # l b)); auto.
    unfold Cf'; apply inc_transitive with (1 := without_sg_cup _ Hcst _ (l a o l b ~~ m)); pred_solve.
   
    (* X (F A * B : m) *)

    destruct Hp as [ Hp1 Hp2 ].
    classic_disj H3 H4.
    generalize (pmo_help_lemma_9 _ _ Hp2); intros ( Hp3 & Hp4 & Hp5 ).
    apply not_fconsistent_CSS in H3; auto.
    apply not_fconsistent_CSS in H4; auto.
    destruct H4 as ( (XfB,CfB) & [ H61 [H62 H63] ] & [H71 H72] & H8 ).
    destruct H3 as ( (XfA,CfA) & [ H31 [H32 H33] ] & [H41 H42] & H5 ).
    destruct (pmo_cpct Hp2) as (C0 & H9 & H10 & H11). 
    generalize (pmo_rt H11); intros H12.
    simpl in H41, H42, H71, H72; cup_tac.
    
    set (Xf' := without_sg XfA (F A # x) cup without_sg XfB (F B # y) cup sg (F A * B # m)).
    set (Cf' := CfA cup CfB cup C0).
    assert (Xf' (F A * B # m)) as G0. unfold Xf'; pred_solve.
    assert (finite Xf') as G2. unfold Xf'; auto.
    assert (finite Cf') as G3. unfold Cf'; auto.
    assert (pmo Cf' (m~~m)) as G4. unfold Cf'; apply pmo_mono with (2 := H12); auto.
    assert (pmo Cf' (x o y ~~ m)) as G8. unfold Cf'; apply pmo_mono with (2 := H11); auto.
    assert (CSS (Xf',Cf')) as G1.
      unfold Xf', Cf'.
      apply CSS_cup_l; auto.
      apply CSS_cup_rl.
      apply CSS_cup_l; auto.
      apply CSS_cup_rl, CSS_dec_l with (2 := H31); auto.
      apply CSS_cup_rr, CSS_dec_l with (2 := H61); auto. 
    assert (CSS_consistent (Xf',Cf')) as G5. unfold  Xf', Cf'; apply H2; auto.
 
    assert(~ CSS_consistent (Xf' cup sg (F A # x), Cf' cup empty)) as G6.
      unfold Xf', Cf'.
      intros H; apply H5.
      apply CSS_consistent_closed with (4 := H); auto.
      apply inc_transitive with (1 := without_sg_cup _ Hstm _ (F A # x)); pred_solve.
      apply inc_transitive with (2 := pmo_inc _); auto; pred_solve.
    
    assert(~ CSS_consistent (Xf' cup sg (F B # y), Cf' cup empty)) as G7.
      unfold Xf', Cf'.
      intros H; apply H8.
      apply CSS_consistent_closed with (4 := H); auto.
      apply inc_transitive with (1 := without_sg_cup _ Hstm _ (F B # y)); pred_solve.
      apply inc_transitive with (2 := pmo_inc _); auto; pred_solve.

    generalize (pmo_help_lemma_9 _ _ G8); intros ( G9 & G10 & _ ).
    destruct fCSS_not_consistent with (2 := G6) as ( tA & HtA1 & HtA2 ); auto.
    destruct fCSS_not_consistent with (2 := G7) as ( tB & HtB1 & HtB2 ); auto.
 
    generalize (bi_tableau_exp_Fmult _ Cf' _ _ _ _ _ G0 G8); intros G11.
    generalize (bi_tab_comp _ nil _ _ _ HtA1 G11); simpl; intros G12.
    generalize (bi_tab_comp _ _ nil _ _ HtB1 G12); simpl; intros G13.
    rewrite <- app_nil_end in G13.
    
    assert (~ CSS_consistent (Xf', Cf')) as G14.
      intros H; apply H.
      exists (tA++tB).
      split; auto.
    apply G14.
    apply H2; auto.
    unfold Xf', Cf'; split; simpl; auto.
  
    (* X (T A -* B : m) *)

    destruct Hp as [ Hp1 Hp2 ].
    classic_disj H3 H4.
    generalize (pmo_help_lemma_9 _ _ Hp2); intros ( Hp3 & Hp4 & Hp5 ).
    apply not_fconsistent_CSS in H3; auto.
    apply not_fconsistent_CSS in H4; auto.
    destruct H4 as ( (XfB,CfB) & [ H61 [H62 H63] ] & [H71 H72] & H8 ).
    destruct H3 as ( (XfA,CfA) & [ H31 [H32 H33] ] & [H41 H42] & H5 ).
    destruct (pmo_cpct Hp2) as (C0 & H9 & H10 & H11).
    generalize (pmo_help_lemma_9 _ _ H11);  intros ( H12 & H13 & H14 ).
    simpl in H41, H42, H71, H72; cup_tac.

    set (Xf' := without_sg XfA (F A # x) cup without_sg XfB (T B # y) cup sg (T A -* B # m)).
    set (Cf' := CfA cup CfB cup C0).
    
    assert (Xf' (T A -* B # m)) as G0. unfold Xf'; pred_solve.
    assert (finite Xf') as G2. unfold Xf'; auto.
    assert (finite Cf') as G3. unfold Cf'; auto.
    assert (pmo Cf' (m~~m)) as G4. unfold Cf'; apply pmo_mono with (2 := H13); auto.
    assert (pmo Cf' (x o m ~~ y)) as G8. unfold Cf'; apply pmo_mono with (2 := H11); auto.
    assert (CSS (Xf',Cf')) as G1.
      unfold Xf', Cf'.
      apply CSS_cup_l; auto.
      apply CSS_cup_rl.
      apply CSS_cup_l; auto.
      apply CSS_cup_rl, CSS_dec_l with (2 := H31); auto.
      apply CSS_cup_rr, CSS_dec_l with (2 := H61); auto. 
    assert (CSS_consistent (Xf',Cf')) as G5. unfold  Xf', Cf'; apply H2; auto.
 
    assert(~ CSS_consistent (Xf' cup sg (F A # x), Cf' cup empty)) as G6.
      unfold Xf', Cf'.
      intros H; apply H5.
      apply CSS_consistent_closed with (4 := H); auto.
      apply inc_transitive with (1 := without_sg_cup _ Hstm _ (F A # x)); pred_solve.
      apply inc_transitive with (2 := pmo_inc _); auto; pred_solve.
    
    assert(~ CSS_consistent (Xf' cup sg (T B # y), Cf' cup empty)) as G7.
      unfold Xf', Cf'.
      intros H; apply H8.
      apply CSS_consistent_closed with (4 := H); auto.
      apply inc_transitive with (1 := without_sg_cup _ Hstm _ (T B # y)); pred_solve.
      apply inc_transitive with (2 := pmo_inc _); auto; pred_solve.

    generalize (pmo_help_lemma_9 _ _ G8); intros ( G9 & _ & G10 ).
    destruct fCSS_not_consistent with (2 := G6) as ( tA & HtA1 & HtA2 ); auto.
    destruct fCSS_not_consistent with (2 := G7) as ( tB & HtB1 & HtB2 ); auto.
    generalize (bi_tableau_exp_Twand _ Cf' _ _ _ _ _ G0 G8); intros G11.
    generalize (bi_tab_comp _ nil _ _ _ HtA1 G11); simpl; intros G12.
    generalize (bi_tab_comp _ _ nil _ _ HtB1 G12); simpl; intros G13.
    rewrite <- app_nil_end in G13.
    
    assert (~ CSS_consistent (Xf', Cf')) as G14.
      intros H; apply H.
      exists (tA++tB).
      split; auto.
    apply G14.
    apply H2; auto.
    unfold Xf', Cf'; split; simpl; auto.

    (* X (F A -* B : m) *)
    
    destruct Hp as ( Hp1 & Hp2 & Hp3 & Hp4 ).
    generalize (H1 _ _ _ Hp1); intros Hp'.
    cutrewrite (a::m = l a o m); auto.
    assert (pmo (C cup sg (l a o m ~~ l b)) (l a ~~ l a)) as Ha.
      apply pmo_help_lemma_2 with (z := l b) (y := m), pmo_inc; pred_solve.
    assert (pmo (C cup sg (l a o m ~~ l b)) (l b ~~ l b)) as Hb.
      apply pmo_help_lemma_4 with (x := l a) (y := m), pmo_inc; pred_solve.

    split; auto. 
    intros (Xf,Cf) [ H3 [ HXf XCf ] ] [ H4 H4' ].
    simpl in H4, H4'.
    destruct (pmo_cpct Hp') as (C0 & H5 & H6 & H7).
    set (Xf' := without_pair Xf (T A # l a) (F B # l b) cup sg (F A -* B # m)).
    assert (CSS (Xf',C)) as H8.
      unfold Xf'.
      apply CSS_dec_l with (2 := H1).
      apply without_pair_inc_l in H4.
      pred_solve.
    destruct CSS_compact with (1 := H8) as ( C1 & H9 & ( H10 & H11 & H12 ) ).
      unfold Xf'; auto.
    set (Cf' := without_sg Cf (l a o m ~~ l b) cup (C1 cup C0)).

    assert (Xf' (F A -* B # m)) as G0. unfold Xf'; pred_solve.
    assert (finite Xf') as G2. unfold Xf'; auto.
    assert (finite Cf') as G3. unfold Cf'; auto.
    assert (CSS (Xf',Cf')) as G1. unfold Cf'; auto.
    assert (pmo Cf' (m~~m)) as G4. unfold Cf'; apply pmo_mono with (2 := H7); pred_solve.    
    assert (CSS_consistent (Xf',Cf')) as G5. unfold  Xf', Cf'; apply H2; auto.
    assert (~ alphabet Cf' a) as Ha'. unfold Cf'; intros H; apply Hp3, alphabet_inc with (2 := H); auto.
    assert (~ alphabet Cf' b) as Hb'. unfold Cf'; intros H; apply Hp4, alphabet_inc with (2 := H); auto.
    assert (CSS_consistent (Xf' cup (sg (T A # l a) cup sg (F B # l b)), Cf' cup sg (l a o m ~~ l b))) as G6.
      split; auto. split; auto.
      apply CSS_cup_l; auto.
      apply CSS_cup_l; auto; apply CSS_sg_l.
      apply pmo_help_lemma_2 with (y := m) (z := l b), pmo_inc; pred_solve.
      apply pmo_help_lemma_4 with (x := l a) (y := m), pmo_inc; pred_solve.
      intros (t & Ht1 & Ht2). 
      generalize (bi_tableau_exp_Fwand _ Cf' _ _ _ _ _ G0 Hp2 Ha' Hb'); intros G6. 
      generalize (bi_tab_comp _ nil nil _ _ Ht1 G6). 
      simpl; rewrite <- app_nil_end; intros G7.
      destruct G5 as [ _ G5 ]; apply G5. 
      exists t. split; auto.
    
    apply CSS_consistent_down_closed with (3 := G6); auto.
    split; simpl; auto.
    unfold Xf'; apply inc_transitive with (1 := without_pair_cup _ Hstm _ (T A # l a) (F B # l b)); auto.
    unfold Cf'; apply inc_transitive with (1 := without_sg_cup _ Hcst _ (l a o m ~~ l b)); pred_solve.
  Qed.

End saturated.

Hint Resolve fconsistent_CSS fconsistent_dwn_closed fconsistent_finite_char fconsistent_open fconsistent_staturated.

Lemma oracle_exists : { P | oracle P /\ CSS_consistent inc P }.
Proof. 
  exists CSS_fconsistent; split.
  split; auto. 
  intros css Hcss.
  apply finite_consistency_prop; auto.
  destruct Hcss; auto.
Qed.
