(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.

Require Import predicate.
Require Import maps.
Require Import cantorbernstein.

Record partial_enumeration {X} (P : X -> Prop) := MkEnum {
  enum_dmap : nat -> X;
  enum_rmap : X -> nat;
  enum_surj : forall x, P x -> enum_dmap (enum_rmap x) = x
}.

Implicit Arguments enum_dmap [ X P ].
Implicit Arguments enum_rmap [ X P ].

Proposition penum_inc {X} (A B : X -> Prop) : A inc B -> partial_enumeration B -> partial_enumeration A.
Proof.
  intros Inc [ dB rB sB ].
  exists dB rB.
  intros; apply sB, Inc; auto.
Qed.

Proposition penum_id {X} (A B : X -> Prop) : A ~ B -> partial_enumeration A -> partial_enumeration B.
Proof. intros [ ? ? ]; apply penum_inc; auto. Qed.

Definition enumeration X := partial_enumeration (fun _ : X => True).

Proposition enumeration_surj X (eX : enumeration X) : surjective (enum_dmap eX).
Proof.
  intros x. destruct eX as [dX rX sX]. simpl. 
  exists (rX x). rewrite sX; auto.
Qed.

Proposition enumeration_inj X (eX :enumeration X) : injective (enum_rmap eX).
Proof.
  intros x x' H. destruct eX as [dX rX sX].
  simpl in H. 
  rewrite <- (sX x), <- (sX x'), -> H; trivial.
Qed.

Proposition enumeration_dec X : enumeration X -> forall x y : X, { x = y } + { x <> y }.
Proof.
  intros [ dmap rmap surj ] x y.
  destruct (eq_nat_dec (rmap x) (rmap y)) as [ H | H ]; [ left | right ].
  rewrite <- surj, <- H; auto; symmetry; auto.
  intros Hx; apply H; rewrite Hx; auto.
Qed.

Record bijection X Y := MkBij {
  bij_dmap : X -> Y;
  bij_rmap : Y -> X;
  bij_inj : forall x, bij_rmap (bij_dmap x) = x;
  bij_surj : forall x, bij_dmap (bij_rmap x) = x
}.

Definition indexation X := bijection nat X.

Notation idx_dmap := (bij_dmap nat).
Notation idx_rmap := (bij_rmap nat).
Notation idx_inj  := (bij_inj  nat).
Notation idx_surj := (bij_surj nat). 
  
Definition enum_from_idx X : indexation X -> enumeration X.
Proof.
  intros HX. exists (idx_dmap _ HX) (idx_rmap _ HX).
  intros x _. apply idx_surj.
Qed.

Proposition indexation_dinj X (iX : indexation X) : injective (idx_dmap _ iX).
Proof.
  intros x1 x2 Hy.
  rewrite <- (idx_inj _ iX), <- Hy, (idx_inj _ iX).
  auto.
Qed.

Proposition indexation_dsurj X (iX : indexation X) : surjective (idx_dmap _ iX).
Proof.
  intros x.
  exists (idx_rmap _ iX x).
  destruct iX; auto.
Qed.

Proposition indexation_rinj X (iX : indexation X) : injective (idx_rmap _ iX).
Proof.
  intros x1 x2 Hy.
  rewrite <- (idx_surj _ iX), <- Hy, (idx_surj _ iX).
  auto.
Qed.

Proposition indexation_rsurj X (iX : indexation X) : surjective (idx_rmap _ iX).
Proof.
  intros x.
  exists (idx_dmap _ iX x).
  destruct iX; auto.
Qed.

Definition enum_from_bij X Y : bijection X Y -> enumeration X -> enumeration Y.
Proof.
  intros [ db rb H1 H2 ] [ dX rX HX ].
  exists (fun n => db (dX n)) (fun y => rX (rb y)).
  intros; rewrite HX, H2; auto.
Qed.

Definition idx_from_bij X Y : bijection X Y -> indexation X -> indexation Y.
Proof.
  intros [ db rb H1 H2 ] [ dX rX H3 H4 ].
  exists (fun n => db (dX n)) (fun y => rX (rb y)).
  intros; rewrite H1; auto.
  intros; rewrite H4; auto.
Qed.

Definition nat_index : indexation nat.
Proof.
  exists (fun x => x) (fun x => x);
  reflexivity.
Qed.

Record isequence X := MkISequence {
  iseq_seq : nat -> X;
  iseq_sze : X -> nat;
  iseq_inc : forall n, iseq_sze (iseq_seq n) < iseq_sze (iseq_seq (S n))
}.

(* for an indexation, there must be an infinite number of elements in X, 
    which is witnessed by a size-increasing of elements of X             
*)

Definition idx_from_enum_iseq X : enumeration X -> isequence X -> indexation X.
Proof.
  intros [ dmap rmap Hdr ] [ iseq isze iinc ].
  assert (forall x, dmap (rmap x) = x) as H.
  intros x; apply Hdr; trivial. clear Hdr.
  set (dm := index_enc _ _ _ H _ _ iinc).
  set (rm := index_dec _ _ _ H _ _ iinc).
  exists rm dm; [ apply index_enc_dec | apply index_dec_enc ].
Qed.

Section DecidableTotalOrder.

  Variable X : Type.
  Variable eX : enumeration X.

  Let i := enum_rmap eX.

  Let inj : injective i.
  Proof. apply enumeration_inj. Qed.

  Definition dto_enum a b := (i a ) <= (i b).
  
  Proposition dto_enum_refl : forall x, dto_enum x x.
  Proof. intros. red; auto. Qed.

  Proposition dto_enum_antisym : forall x y, dto_enum x y -> dto_enum y x -> x = y.
  Proof. unfold dto_enum. intros. apply inj. apply le_antisym; auto. Qed.

  Proposition dto_enum_trans : forall x y z, dto_enum x y -> dto_enum y z -> dto_enum x z.
  Proof. unfold dto_enum; intros ? ? ?; apply le_trans. Qed.

  Proposition dto_enum_dec : forall x y,  { dto_enum x y } + { ~ dto_enum x y}.
  Proof. intros x y. unfold dto_enum. apply le_dec. Qed.

  Proposition dto_enum_total : forall x y, { dto_enum x y } + { dto_enum y x }.
  Proof. intros. unfold dto_enum. destruct (le_ge_dec (i x) (i y)); tauto. Qed.

End DecidableTotalOrder.



