(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import Arith.
Require Import MinMax.
Require Import List.

Require Import predicate.
Require Import maps.
Require Import natseq.
Require Import enum.
Require Import nat2.
Require Import maxf.

Require Import dlist.
Require Import term.

Section NatRepeat.

  Definition nat_repeat n : nat := snd (idx_dmap _ nat2_index n).

  Let nr_inv j i := idx_rmap _ nat2_index (i,j).

  Let nr_inv_inj j : fnat_inj (nr_inv j).
  Proof.
    intros i i' H.
    assert ((i,j)=(i',j)) as H'.
      rewrite <- (idx_surj _ nat2_index (i,j)); trivial.
      rewrite <- (idx_surj _ nat2_index (i',j)); trivial.
      unfold nr_inv in H. rewrite -> H. reflexivity.
    injection H'; trivial.
  Qed.

  Proposition nat_repeat_isurj : i_surjective nat_repeat.
  Proof.
    intros x n.
    destruct (fnat_inj_prop _ (nr_inv_inj x) n) as [ k  [ Hk1 Hk2 ] ].
    exists (nr_inv x k). 
    split; auto.
    unfold r_image, nat_repeat, nr_inv.
    rewrite -> (idx_surj _ nat2_index); trivial.
  Qed.

End NatRepeat.

Definition unit_enum : enumeration unit.
Proof.
  exists (fun _ => tt) (fun _ => 0).
  intros []; auto.
Qed.

Definition bool_enum : enumeration bool.
Proof.
  set (dB := fun c => match c with O => false | _ => true end).
  set (rB := fun c => match c with false => O | true => 1 end).
  exists dB rB.
  intros c H. induction c; simpl; reflexivity.
Qed.

Definition nat_enum : enumeration nat.
Proof. apply (enum_from_idx _ nat_index). Qed.

Definition pred_prod {X Y} (P : X -> Prop) (Q : Y -> Prop) c := match c with (x,y) => P x /\ Q y end.

Definition penum_prod {X Y} {P : X -> Prop} {Q : Y -> Prop} 
                           (eP : partial_enumeration P) 
                           (eQ : partial_enumeration Q) 
                          : partial_enumeration (pred_prod P Q).
Proof.
  destruct eP as [dX rX sX].
  destruct eQ as [dY rY sY].
  destruct nat2_index as [dN rN _ sN ].
  set (dP := fun c => match dN c with (a,b) => (dX a,dY b) end).
  set (rP := fun c => match c with (x,y) => rN (rX x, rY y) end).
  exists dP rP.
  intros [x y] [ ? ? ].
  unfold dP, rP. 
  rewrite -> sN, -> sX; auto.
  rewrite -> sY; auto.
Qed.

Definition enum_prod {X Y} : enumeration X -> enumeration Y -> enumeration (X*Y).
Proof.
  intros eX eY.  
  destruct (penum_prod eX eY) as [ dZ rZ HZ ].
  exists dZ rZ.
  intros [] _; rewrite (HZ _); simpl; auto. 
Qed.

Definition pred_sum {X Y} (P : X -> Prop) (Q : Y -> Prop) c := match c with inl x => P x | inr y => Q y end.

Definition penum_sum {X Y} {P : X -> Prop} {Q : Y -> Prop} 
                           (eP : partial_enumeration P) 
                           (eQ : partial_enumeration Q) 
                          : partial_enumeration (pred_sum P Q).
Proof.
  destruct eP as [dX rX sX].
  destruct eQ as [dY rY sY].
  destruct nat2_index as [dN rN _ sN ].
  set (dP := fun c => match dN c with (0,a) => inl _ (dX a) | (_,b) => inr _ (dY b) end).
  set (rP := fun c => match c with inl x => rN (0, rX x) | inr y => rN (1, rY y) end).
  exists dP rP.
  intros [ x | y ] H; unfold dP; unfold rP; rewrite -> sN; trivial. 
  rewrite -> sX; trivial. rewrite -> sY; trivial.
Qed.

Definition enum_sum {X Y} : enumeration X -> enumeration Y -> enumeration (X+Y).
Proof.
  intros eX eY.  
  destruct (penum_sum eX eY) as [ dZ rZ HZ ].
  exists dZ rZ.
  intros [] _; rewrite (HZ _); simpl; auto. 
Qed.

Definition enum_union {X} {phi : nat -> X -> Prop} {psy : X -> nat} 
                   (t : forall x, (union phi) x -> phi (psy x) x) 
                   (e : forall n, partial_enumeration (phi n)) : partial_enumeration (union phi).
Proof.
  destruct nat2_index as [dm rm _ srj].
  set (dmap0 := fun n => match dm n with (a,b) => enum_dmap (e a) b end).
  set (rmap0 := fun x => rm (psy x, enum_rmap (e (psy x)) x)).
  exists dmap0 rmap0.
  intros; unfold dmap0, rmap0.
  rewrite -> srj; trivial.
  destruct (e (psy x)) as [dm1 sm1 srj1]. simpl.
  apply srj1. apply t. trivial.
Qed.

Definition enum_layers X {phi : nat -> X -> Prop} {psy : X -> nat} 
                                              (t : forall x, phi (psy x) x) 
                                              (e : forall n, partial_enumeration (phi n)) 
                                            : enumeration X.
Proof.
  apply (penum_id (union phi)).
  split; intros x Hx. trivial. exists (psy x). apply t.
  apply (@enum_union _ _ psy); trivial.
Qed.

(* senum X : X is either enumerable or empty *)

Definition senum X := enumeration (unit+X).

Definition empty_senum : senum Empty_set.
Proof.
  exists (fun _ => inl _ tt) (fun _ => 0).
  intros [ [] | [] ] _; auto.
Qed.

Definition unit_senum : senum unit.
Proof. apply enum_sum; apply unit_enum. Qed.

Proposition ssenum_senum {X} : senum (unit + X) -> senum X.
Proof.
  intros [ dX rX HX ].
  exists (fun n => match dX n with
                     | inl _       => inl _ tt
                     | inr (inl _) => inl _ tt
                     | inr (inr x) => inr _ x
                   end)
         (fun c => rX (inr _ c)).
  intros [ [] | x ] _; rewrite HX; auto.
Qed.

Definition lift_prop {X} (P : X -> Prop) c :=
  match c with
    | inl tt => True
    | inr x  => P x
  end.  

Theorem spenum_penum {X} {P : X -> Prop} x : P x -> partial_enumeration (lift_prop P) -> partial_enumeration P.
Proof.
  intros Hx [ dX rX HX ].
  exists (fun n => match dX n with inl _ => x | inr y => y end)
         (fun y => rX (inr _ y)).
  intros; rewrite HX; auto.
Qed.

Theorem spenum_prod {X} {Y}  {P : X -> Prop} {Q : Y -> Prop} : 
           partial_enumeration (lift_prop P)
        -> partial_enumeration (lift_prop Q)
        -> partial_enumeration (lift_prop (pred_prod P Q)).
Proof.
  intros [ dX rX HX ] [ dY rY HY ].
  destruct nat2_index as [dN rN _ HN ].
  set (dZ n := match dN n with (a,b) => 
                 match (dX a, dY b) with
                   | (inl _, _    ) => inl _ tt
                   | (_    , inl _) => inl _ tt
                   | (inr x, inr y) => inr _ (x,y)
                 end
               end).
  set (rZ c := match c with
                 | inl tt    => rN (rX (inl _ tt), rY (inl _ tt))
                 | inr (x,y) => rN (rX (inr _ x), rY (inr _ y))
               end).
  exists dZ rZ.
  intros [ [] | (x,y) ] H; unfold rZ, dZ; rewrite HN, HX; auto; simpl in H |- *;
  try (rewrite HY; simpl); tauto.
Qed.

Theorem spenum_sum {X} {Y}  {P : X -> Prop} {Q : Y -> Prop} : 
           partial_enumeration (lift_prop P)
        -> partial_enumeration (lift_prop Q)
        -> partial_enumeration (lift_prop (pred_sum P Q)).
Proof.
  intros [ dX rX HX ] [ dY rY HY ].
  destruct (enum_sum unit_enum (enum_sum nat_enum nat_enum)) as [dN rN HN ].
  set (dZ n := match dN n with 
                 | inl _       => inl _ tt
                 | inr (inl a) => match dX a with 
                                    | inl _ => inl _ tt
                                    | inr x => inr _ (inl _ x)
                                  end
                 | inr (inr b) => match dY b with 
                                    | inl _ => inl _ tt
                                    | inr y => inr _ (inr _ y)
                                  end
               end).
  set (rZ (c : unit + (X + Y)) := rN (match c with
                     | inl _       => inl _ tt 
                     | inr (inl x) => inr _ (inl _ (rX (inr _ x)))
                     | inr (inr y) => inr _ (inr _ (rY (inr _ y)))
                   end)).
  exists dZ rZ.
  intros [ [] | [ x | y ] ]; unfold rZ, dZ; rewrite HN; simpl; auto.
  intros; rewrite HX; simpl; auto.
  intros; rewrite HY; simpl; auto.
Qed.

Theorem penum_spenum {X} {P : X -> Prop} : partial_enumeration P -> partial_enumeration (lift_prop P) .
Proof.
  intros [ dX rX HX ].
  set (dZ n := match n with
                 | 0   => inl _ tt
                 | S n => inr _ (dX n)
               end).
  set (rZ (c : unit + X) := match c with
                 | inl _ => 0
                 | inr x => S (rX x)
               end).
  exists dZ rZ.
  intros [ [] | x ] H; simpl; auto.
  rewrite HX; auto.
Qed.

(* senum X <-> enumeration X /\ X *)

Theorem senum_enum {X} : X -> senum X -> enumeration X.
Proof.
  intros x [ ds rs Hs ].
  set (dt n := match ds n with inl _ => x | inr y => y end).
  set (rt y := rs (inr _ y)).
  exists dt rt.
  intros y _; unfold dt, rt.
  rewrite Hs; auto.
Qed.

Theorem enum_senum {X} : enumeration X -> senum X.
Proof.
  intros.
  apply enum_sum; auto.
  apply unit_enum.
Qed.

Theorem enum_element {X} : enumeration X -> X.
Proof. intros [ ds ? ? ]; apply (ds 0); auto. Qed.

Hint Resolve @senum_enum @enum_senum.

Definition pred_isum {t : nat -> Type} (P : forall n, t n -> Prop) (c : sigT t) := match c with existT n x => P n x end.

Section penum_isum.

  Variable t : nat -> Type.
  Variable P : forall n, t n -> Prop.
  Variable st : forall n, partial_enumeration (lift_prop (P n)).

  Let eX := enum_sum unit_enum (enum_prod nat_enum nat_enum).
  Let dX := enum_dmap eX.
  Let rX := enum_rmap eX.
  Let HX x : dX (rX x) = x.
  Proof. apply enum_surj; auto. Qed.

  Let dp n := 
    match dX n with 
      | inl _     => inl _ tt
      | inr (i,j) => match enum_dmap (st i) j with 
                       | inl _ => inl _ tt
                       | inr x => inr _ (existT t i x)
                     end
    end.

  Let rp (c : unit + sigT t) :=
    match c with
      | inl _ => rX (inl _ tt)
      | inr (existT i c) => rX (inr _ (i,(enum_rmap (st i) (inr _ c))))
    end.

  Theorem penum_isum : partial_enumeration (lift_prop (pred_isum P)).
  Proof.
    exists dp rp.
    intros [ [] | [ i x ] ] H;
    unfold rp, dp; rewrite HX; auto.
    rewrite enum_surj; auto.
  Qed.

End penum_isum.

Section dlist_spenum.

  Variable A : Type.
  Variable P : A -> Prop.
  Variable eA : partial_enumeration (lift_prop P).

  Theorem dlist_spenum : forall n, partial_enumeration (lift_prop (dl_fall P (n:=n))). 
  Proof.
    destruct eA as [dA rA sA].
    induction n.
    
    set (dL n := match n with 
                   | 0 => inl _ tt
                   | _ => inr _ (dl_nil : dlist A 0)
                 end).
    set (rL (c : unit+dlist A 0) := 
                 match c with
                   | inl tt => 0
                   | inr _  => 1
                 end).
    exists dL rL.
    intros [[] | ]; cbv; symmetry; auto. 
    f_equal; apply dl_0_nil.

    destruct (spenum_prod eA IHn) as [dTn rTn HTn].
    set (dZ n := match dTn n with
                   | inl _     => inl _ tt
                   | inr (a,l) => inr _ (a##l)
                 end).
    set (rZ (c : unit+dlist A (S n)) := rTn (match c with
                   | inl _ => inl _ tt
                   | inr l => inr _ (dl_head l,dl_tail l)
                 end)).
    exists dZ rZ.
    intros [ [] | l ] H; unfold rZ, dZ; rewrite HTn; auto.
    rewrite (dl_head_tail l); simpl; auto.
    simpl in H |- *.
    apply dl_fall_cons; auto.
    rewrite <- (dl_head_tail l); auto.    
  Qed.

End dlist_spenum.

Section dlist_penum.

  Variable A : Type.
  Variable P : A -> Prop.
  Variable a : A.
  Hypothesis Ha : P a.
  Variable eA : partial_enumeration P.

  Let an := fix loop n := 
              match n  return dlist A n with 
                | 0   => dl_nil
                | S n => a##(loop n)
              end.

  Theorem dlist_penum : forall n, partial_enumeration (dl_fall P (n := n)). 
  Proof.
    intros n.    
    apply spenum_penum with (x := an n).
    induction n; simpl.
    apply dl_fall_nil.
    apply dl_fall_cons; auto.
    apply dlist_spenum.
    apply penum_spenum.
    auto.
  Qed.

End dlist_penum.

Implicit Arguments dlist_penum [ A P ].

Theorem dlist_enum {X} : enumeration X -> forall n, enumeration (dlist X n).
Proof.
  intros eX n.
  set (x := enum_dmap eX 0).
  assert (Hx : (fun _ => True) x). auto.
  destruct (dlist_penum _ Hx eX n) as [ dn rn Hn ].
  exists dn rn; intros; rewrite Hn; auto.
  intro; auto.
Qed.

(* We can enumerate terms over any enumerable signature *)

Section Term_senum.

  Variable sig : nat -> Type.

  Hypothesis sig_enum : forall n, senum (sig n).

  Let ht (x : unit + term sig) := 
    match x with 
      | inl tt => 0
      | inr t  => term_height t
    end.

  Let term_enum_rec h : partial_enumeration (fun x => ht x <= h).
  Proof.
    induction h as [ | h ].
     
    set (dL (_: nat) := inl (term sig) tt).
    set (rL (_ : unit + term sig) := 0).
    exists dL rL.
    intros [ [] | [] ] H; unfold rL, dL; auto.
    simpl in H; contradiction (le_Sn_O _ H).

    assert (partial_enumeration (lift_prop (fun (t : term sig) => term_height t <= h))) as IH.
      apply penum_inc with (2 := IHh).
      intros [[] | x ]; simpl; auto.
      intros _; apply le_O_n.
    clear IHh.
    
    assert (forall n, partial_enumeration (lift_prop (pred_prod (fun _ : sig n => True) (@dl_fall _ (fun t : term sig => term_height t <= h) n)))) as H1.
      intros n.  
      apply spenum_prod; auto.
      apply penum_inc with (2 := sig_enum n).
      intros [ [] | ? ]; auto.
      apply dlist_spenum; auto.
      
    destruct (penum_isum _ _ H1) as [ dX rX HX ]; clear IH H1.
    
    set (dZ n := match dX n with
                   | inl _                => inl _ tt
                   | inr (existT n (f,v)) => inr _ (f[v])
                 end).
    set (rZ (c : unit + term sig ) := rX (match c with 
                   | inl _               => inl _ tt
                   | inr (in_term n f v) => inr _ (existT _ n (f,v))
                 end)).
    exists dZ rZ.
    intros [ [] | [ n f v ] ] H; unfold rZ, dZ; rewrite HX; simpl in |- *; auto.
    split; auto.
    unfold ht in H; rewrite term_height_prop in H.
    apply le_S_n in H.
    intros x Hx.
    apply le_trans with (2 := H).
    rewrite <- dl_get_map.
    apply dl_get_max.
  Qed.
 
  Theorem term_senum : senum (term sig).
  Proof.
    unfold senum.
    apply enum_layers with (psy := ht) (2 := term_enum_rec).
    intros; auto.
  Qed.

End Term_senum.

Section Term_empty.

  Variable sig : nat -> Type.

  (* when there is no constant (arity 0), the term type is empty *)

  Hypothesis Hsig : sig 0 -> False.

  Theorem term_empty : term sig -> False.
  Proof.
    induction 1 as [ n c v IH ].
    destruct n.
    apply Hsig, c.
    rewrite (dl_head_tail v) in IH.
    apply dl_fall_cons in IH; destruct IH; auto.
  Qed.

End Term_empty.

Section Term_enum_idx.

  Variable sig : nat -> Type.
  Variable sig_enum : forall n, senum (sig n).

  Variable c0 : sig 0.

  (* when there is a constant (arity 0), we can enumerate term type *)

  Theorem term_enum : enumeration (term sig).
  Proof. 
    apply senum_enum.
    apply (in_term c0 dl_nil).
    apply term_senum; auto.
  Qed.

  Variable n : nat.
  Variable cn : sig (S n).

  (* we suppose there exists at least a constant (arity 0) and at least a function symbol (arity > 0) *)

  (* We have a height increasing sequence of terms *)

  Let Th h : { t : term sig | term_height t = S h }.
  Proof.
    induction h as [ | h IH ].
    exists (@in_term _ 0 c0 dl_nil); auto.
    destruct IH as [ t Ht ].
    set (v0 := dl_repeat n (c0[dl_nil])).
    assert (dl_max (dl_map (@term_height _) v0) <= 1) as Hv0.
      apply dl_max_lub.
      intros x Hx.
      unfold v0.
      rewrite dl_get_map, dl_repeat_prop; auto.
    exists (cn[t##v0]).
    rewrite term_height_prop.
    f_equal.
    simpl.
    rewrite Ht.
    apply max_l, le_trans with (1 := Hv0).
    omega.
  Qed.

  Theorem term_index : indexation (term sig).
  Proof.
    apply idx_from_enum_iseq; auto.
    apply term_enum; auto.
    exists (fun h => projT1 (Th h)) (@term_height sig).
    intros h; do 2 rewrite (projT2 (Th _)); omega.
  Qed.

End Term_enum_idx.

Implicit Arguments term_enum [ sig ].
Implicit Arguments term_index [ sig ].

Section List_idx.

  Variable X : Type.

  Let sig n : Type := match n with 
                 | 0 => unit
                 | 1 => X
                 | _ => Empty_set
               end.

  Let f n := match n as k return sig k -> dlist (list X) k -> list X with
               | 0 => fun tt _ => @nil X
               | 1 => fun x  v  => x::(dl_head v)
               | _ => fun c  _ => Empty_set_rect (fun _ => list X) c
             end.

  Let term_list := term_recX f.

  Let term_list_prop n (c : sig n) v : term_list (c[v]) = f n c (dl_map term_list v).
  Proof. apply term_recX_prop. Qed.

  Let list_term : list X -> term sig :=
    (fix loop (l : list X) := 
    match l return term sig with 
      | nil    => tt[dl_nil]
      | x::l   => @in_term sig 1 x (loop l##dl_nil)
    end).

  Let term_list_eq l : term_list (list_term l) = l.
  Proof.
    induction l; simpl; auto.
    rewrite IHl; auto.
  Qed.

  Let list_term_eq t : list_term (term_list t) = t.
  Proof.
    induction t as [ n c v Hv ].
    rewrite term_list_prop.
    destruct n.
    destruct c.
    simpl.
    rewrite (dl_0_nil v); auto.
    destruct n.
    simpl in c |- *.
    f_equal.
    rewrite (dl_head_tail v); simpl.
    rewrite dl_head_tail in Hv.
    apply dl_fall_cons in Hv.
    destruct Hv as [ H1 H2 ].
    rewrite H1.
    f_equal.
    rewrite (dl_0_nil (dl_tail v)); auto.
    simpl in c.
    destruct c.
  Qed.

  (* we have a bijection between terms over sig and bin_tree *)

  Let list_term_bij : bijection (term sig) (list X).
  Proof. exists term_list list_term; auto. Qed.

  Variable eX : enumeration X.

  Hint Resolve unit_senum empty_senum.

  Theorem list_index : indexation (list X).
  Proof.
    apply idx_from_bij with (1:= list_term_bij).
    apply term_index with (n := 0).
    destruct n; auto.
    destruct n; auto.
    apply tt.
    apply enum_element; auto.
  Qed.

  Hint Resolve list_index.

  Theorem list_enum : enumeration (list X).
  Proof. apply enum_from_idx; auto. Qed.
    
End List_idx.

Implicit Arguments list_enum [ X ].
Implicit Arguments list_index [ X ].

Section Tree_idx.

  Variable X : Type. 

  Inductive bin_tree : Type :=
    | leaf : X -> bin_tree
    | node : bin_tree -> bin_tree -> bin_tree.

  Let sig n : Type := match n with 
                 | 0 => X
                 | 2 => unit
                 | _ => Empty_set
               end.

  Let f n := match n as k return sig k -> dlist bin_tree k -> bin_tree with
               | 0 => fun x  _ => leaf x 
               | 2 => fun tt v => node (dl_head v) (dl_head (dl_tail v))
               | _ => fun c  _ => Empty_set_rect (fun _ => bin_tree) c
             end.

  Let term_btree := term_recX f.

  Let term_btree_prop n (c : sig n) v : term_btree (c[v]) = f n c (dl_map term_btree v).
  Proof. apply term_recX_prop. Qed.

  Let btree_term : bin_tree -> term sig :=
    (fix loop t := 
    match t return term sig with 
      | leaf x     => @in_term sig 0 x dl_nil
      | node t1 t2 => @in_term sig 2 tt (loop t1##loop t2##dl_nil)
    end).

  Let term_btree_eq t : term_btree (btree_term t) = t.
  Proof.
    induction t; simpl; auto.
    rewrite IHt1, IHt2; auto.
  Qed.

  Let btree_term_eq t : btree_term (term_btree t) = t.
  Proof.
    induction t as [ n c v Hv ].
    rewrite term_btree_prop.
    destruct n.
    simpl in c |- *.
    rewrite (dl_0_nil v); auto.
    destruct n.
    destruct c.
    destruct n.
    destruct c.
    simpl.
    rewrite dl_head_tail in Hv.
    rewrite (dl_head_tail (dl_tail v)) in Hv.
    apply dl_fall_cons in Hv.
    destruct Hv as [ H1 H2 ].
    apply dl_fall_cons in H2.
    destruct H2 as [ H2 _ ].
    rewrite (dl_head_tail v), (dl_head_tail (dl_tail v)).
    simpl dl_map.
    simpl dl_head.
    rewrite H1, H2, (dl_0_nil (dl_tail _)); auto.
    destruct c.
  Qed.

  (* we have a bijection between terms over sig and bin_tree *)

  Let btree_term_bij : bijection (term sig) bin_tree.
  Proof. exists term_btree btree_term; auto. Qed.
 
  Variable eX : enumeration X.

  Hint Resolve unit_senum empty_senum.

  Theorem bin_tree_index : indexation bin_tree.
  Proof.
    apply idx_from_bij with (1:= btree_term_bij).    
    apply term_index with (n := 1).
    destruct n.
    apply enum_senum; auto.
    destruct n; auto.
    destruct n; auto.
    apply enum_element; auto.
    apply tt.
  Qed.

  Hint Resolve bin_tree_index.

  Theorem bin_tree_enum : enumeration bin_tree.
  Proof. apply enum_from_idx; auto. Qed.

End Tree_idx.

Implicit Arguments bin_tree_enum [ X ].
Implicit Arguments bin_tree_index [ X ].

Definition nat_bt_surj := enum_dmap (bin_tree_enum nat_enum).

Proposition bt_surj : surjective nat_bt_surj.
Proof. apply enumeration_surj. Qed.


