(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import predicate.
Require Import carrier.
Require Import branch.

Section Tableau.

  Variable branch : Type.
  Variable beq : branch -> branch -> Prop.
  
  Infix "eq" := beq (at level 60).

  Hypothesis eq_refl : forall x, x eq x.
  Hypothesis eq_sym  : forall x y, x eq y -> y eq x.
  Hypothesis eq_tran : forall x y z, x eq y -> y eq z -> x eq z.
  
  Infix "leq" := (listbranch_eq beq) (at level 60).

  Proposition listbranch_eq_refl l : l leq l.
  Proof. induction l; simpl; auto. Qed.
  
  Hint Immediate listbranch_eq_refl.
  
  Proposition listbranch_eq_sym l : forall m, l leq m -> m leq l.
  Proof.
    induction l; induction m; simpl; auto.
    intros [H1 H2]; simpl in IHm.
    split; auto.
  Qed.
  
  Hint Immediate listbranch_eq_sym.
  
  Proposition listbranch_eq_tran l m n : l leq m -> m leq n -> l leq n.
  Proof.
    intros H1 H2; generalize m H1 n H2; clear m n H1 H2.
    induction l; induction m; intros H1; induction n; intros H2; simpl; auto.
    contradict H2; auto.
    simpl in H1; destruct H1.
    simpl in H2; destruct H2.
    split. 
    apply eq_tran with (1 := H); auto.
    apply IHl with (2 := H2); auto.
  Qed.

  (* stability for P: if P hold for b then it holds for one of its expansions, up to equivalence *)

  Variable expansion : branch -> list branch -> Prop.

  Definition expansion_stable (P : branch -> Prop) := forall b lb, expansion b lb -> P b -> exists x, exists y, carrier lb y /\ x eq y /\ P x.

  Inductive tableau (b0 : branch) : list branch -> Prop :=
    | In_tab0 : forall b1, b0 eq b1 -> tableau b0 (b1::nil)
    | In_tab1 : forall l1 b l2 lb l', expansion b lb -> (l1++lb++l2) leq l' -> tableau b0 (l1++b::l2) -> tableau b0 l'.

  Hint Constructors tableau.

  (* the tableau predicate is closed under equivalence (even if expansion is not) *)

  Proposition tableau_eq b0 b1 t0 t1 : b0 eq b1 -> t0 leq t1 -> tableau b0 t0 -> tableau b1 t1.
  Proof.
    intros H0 H1 H2; generalize b0 t0 H2 b1 t1 H0 H1; clear b0 b1 t0 t1 H0 H1 H2.
    induction 1 as [ b1 Hb1 | l1 b l2 lb l' ]; intros b2 t2 H11 H12.
    destruct t2 as [ | b3 [ | ] ]; simpl in H12; try tauto.
    destruct H12 as [ H12 _ ].
    constructor 1.
    apply eq_tran with (2 := H12),
          eq_tran with (2 := Hb1), eq_sym; auto.
    constructor 2 with (1 := H) (l1 := l1) (l2 := l2).
    apply listbranch_eq_tran with (2 := H12); auto.
    apply IHtableau; auto.
  Qed.

  Proposition tableau_start b : tableau b (b::nil).
  Proof. constructor; auto. Qed.

  Proposition tableau_next b0 l1 b l2 lb : expansion b lb -> tableau b0 (l1++b::l2) -> tableau b0 (l1++lb++l2).
  Proof. intros H1 H2; constructor 2 with (1 := H1) (3 := H2); auto. Qed.
  
  Proposition tableau_comp b0 l1 l2 b t :  tableau b t 
                                        -> tableau b0 (l1 ++ b :: l2)
                                        -> tableau b0 (l1 ++ t ++ l2).
  Proof.
    intros H1 H2; generalize t H1; clear t H1.
    induction 1.

    apply tableau_eq with (3 := H2); auto.
    simpl; apply listbranch_eq_app; simpl; auto.
    
    apply tableau_eq with (b0 := b0) (t0 := (l1++(l0++lb++l3)++l2)); auto.
    cutrewrite (l1 ++ (l0 ++ lb ++ l3) ++ l2 = (l1 ++ l0) ++ lb ++ (l3 ++ l2)).
    cutrewrite (l1 ++ (l0 ++ b1 :: l3) ++ l2 = (l1 ++ l0) ++ b1 :: (l3 ++ l2)) in IHtableau.
    apply tableau_next with (2 := IHtableau); auto.
    simpl; do 2 (rewrite <- ass_app; auto).
    simpl; do 3 (rewrite <- ass_app; auto).
  Qed.

  Section TabExpFun.

    Variable f : branch -> branch.
    Hypothesis f_eq : forall x y, x eq y -> f x eq f y.
    Hypothesis f_exp : forall b lb, expansion b lb -> exists lb', expansion (f b) lb' /\ map f lb leq lb'.

    Let f_leq : forall l1 l2, l1 leq l2 -> map f l1 leq map f l2.
    Proof.
      induction l1; induction l2; simpl; auto; intros [ ]; split; auto.
    Qed.

    Proposition tableau_prop_fun : forall b t, tableau b t -> tableau (f b) (map f t).
    Proof.
      induction 1 as [ b1 Hb1 | l1 b1 l2 lb l' H1 H2 H3 H4 ].
      simpl.
      apply f_eq in Hb1.
      apply eq_sym in Hb1.
      apply tableau_eq with (1 := Hb1) (t0 := f b1::nil); auto.
      destruct (f_exp _ _ H1) as ( b2 & H5 & H6 ).
      apply f_leq in H2.
      apply tableau_eq with (b0 := f b) (t0 := map f l1 ++ b2 ++ map f l2); auto.
      apply listbranch_eq_tran with (2 := H2).
      do 2 rewrite map_app; auto. 
      rewrite map_app in H4; simpl in H4.
      apply tableau_next with (2 := H4); auto.
    Qed.

  End TabExpFun.

  Section TableauPropConj_List.

    Variable b0 : branch.

    Variable P : list branch -> Prop.
    Hypothesis P0 : P (b0::nil).
    Hypothesis P_eq   : forall x y, x eq y -> P (x::nil) -> P (y::nil).
    Hypothesis P_app1 : forall l1 l2, P l1 -> P l2 -> P (l1++l2).
    Hypothesis P_app2 : forall l1 l2, P (l1++l2) -> P l1 /\ P l2.
    Hypothesis P_exp : forall b lb, expansion b lb -> P (b::nil) -> P lb.

    Let P_leq : forall l1 l2, l1 leq l2 -> P l1 -> P l2.
    Proof.
      induction l1; destruct l2; auto.
      intros H; contradiction H.
      intros H; contradiction H.
      intros [ H1 H2 ].
      cutrewrite (a::l1 = (a::nil)++l1); auto.
      cutrewrite (b::l2 = (b::nil)++l2); auto.
      intros H. apply P_app2 in H.
      apply P_app1; auto.
      apply P_eq with (1 := H1); tauto.
      apply IHl1; tauto.
    Qed.

    Proposition tableau_prop_conj_list : forall t, tableau b0 t -> P t.
    Proof.
      induction 1 as [ b1 H1 | l1 b l2 lb l' He Ht IH ].
      apply P_eq with (1 := H1); auto.
      apply P_app2 in IHIH; destruct IHIH as [ IH1 IHIH ].
      cutrewrite (b::l2 = (b::nil)++l2) in IHIH; auto.
      apply P_app2 in IHIH; destruct IHIH as [ IHb IH2 ].
      apply P_leq with (1 := Ht).
      do 2 (apply P_app1; auto).
      apply P_exp with (1 := He); auto.
    Qed.

  End TableauPropConj_List.

  Section TableauPropConj.

    Variable b0 : branch.
    Variable P : branch -> Prop.
    Hypothesis P0 : P b0.
    Hypothesis P_eq : forall x y, x eq y -> P x -> P y. 
    Hypothesis P_exp : forall b lb, P b -> expansion b lb -> carrier lb inc P.

    Proposition tableau_prop_conj : forall t, tableau b0 t -> carrier t inc P.
    Proof.
      apply (tableau_prop_conj_list b0 (fun lb => carrier lb inc P)); auto.
      intros x y H1 H2; apply carrier_cons_2; auto; apply P_eq with (1 := H1); auto.
      intros ? ?; apply carrier_app_1.
      intros b ? ? ?; apply P_exp with (b := b); auto.
    Qed.

  End TableauPropConj.
        
  Section TabPropDisj_List.

    Variable b0 : branch.

    Variable P : list branch -> Prop.
    Hypothesis P0 : P (b0::nil).
    Hypothesis P_eq    : forall x y, x eq y -> P (x::nil) -> P (y::nil).
    Hypothesis P_app_1 : forall l1 l2 l3, P l2 -> P (l1++l2++l3).
    Hypothesis P_app_2 : forall l1 l2, P (l1++l2) -> P l1 \/ P l2. 
    Hypothesis P_exp : forall b lb, P (b::nil) -> expansion b lb -> P lb.

    Let P_leq : forall l1 l2, l1 leq l2 -> P l1 -> P l2.
    Proof.
      induction l1; destruct l2; auto.
      intros H; contradiction H.
      intros H; contradiction H.
      simpl; intros [ H1 H2 ].
      cutrewrite (a::l1 = (a::nil)++l1); auto.
      cutrewrite (b::l2 = (b::nil)++l2); auto.
      intros H. apply P_app_2 in H.
      destruct H as [ H | H ].
      apply P_eq with (1 := H1) in H.
      apply (P_app_1 nil (b::nil) l2); auto.
      apply (IHl1 _ H2) in H.
      cutrewrite ((b::nil)++l2 = (b::nil)++l2++nil).
      apply (P_app_1 (b::nil) l2 nil H). 
      rewrite <- app_nil_end; auto.
    Qed.

    Proposition tableau_prop_disj_rec : forall t, tableau b0 t -> P t.
    Proof.
      induction 1 as [ b1 H2 | l1 b l2 lb l' He Ht IH ].
      apply P_eq with (1 := H2); auto.
      apply P_app_2 in IHIH; destruct IHIH as [ H1 | IHIH ]; apply P_leq with (1 := Ht).
      cutrewrite (l1++lb++l2 = nil++l1++(lb++l2)); auto.
      cutrewrite (b::l2 = (b::nil)++l2) in IHIH; auto.
      apply P_app_2 in IHIH; destruct IHIH as [ Hb | H2 ].
      apply (P_exp _ lb) in Hb; auto.
      cutrewrite (l1++lb++l2 = (l1++lb)++l2++nil); auto.
      rewrite <- app_ass, <- app_nil_end; auto.
    Qed.

  End TabPropDisj_List.

  Section TabExpInc.

    Variable R : branch -> branch -> Prop.
    Hypothesis R_eq : forall x y x' y', x eq x' -> y eq y' -> R x y -> R x' y'.
    Hypothesis R_exp : forall b b' lb, R b b' -> expansion b lb -> exists lb', expansion b' lb' /\ listbranch_eq R lb lb'.

    Let R_leq_1 : forall l1 l2 l1', l1 leq l1' -> listbranch_eq R l1 l2 -> listbranch_eq R l1' l2.
    Proof.
      induction l1; induction l1'; induction l2; simpl; auto; intros H1 H2; 
      try (contradict H1; auto; fail).
      destruct H2 as [ H3 H4 ]; destruct H1 as [ H1 H2 ].
      split.
      apply R_eq with (3 := H3); auto.
      apply IHl1 with (2 := H4); auto.
    Qed.

    Proposition tableau_prop_inc : forall b t, tableau b t -> forall b', R b b' -> exists t', tableau b' t' /\ listbranch_eq R t t'.
    Proof.
      induction 1 as [ b1 Hb1 | l1 b1 l2 lb l' ]; intros b' H'.
      exists (b'::nil).
      split.
      apply tableau_start.
      split; auto.
      apply R_eq with (3 := H'); auto.
      simpl; auto.
      destruct (IHtableau _ H') as ( t & H2 & H3 ).
      apply listbranch_eq_split in H3.
      destruct H3 as ( l1' & b2 & l2' & H3 & H4 & H5 & H6 ).
      destruct (R_exp _ _ _ H4 H) as ( t' & H7 & H8 ).
      exists (l1'++t'++l2').
      split.
      apply tableau_next with (1 := H7).
      rewrite <- H6; auto.
      apply R_leq_1 with (1 := H0); auto.
    Qed.
      
  End TabExpInc.

  Section TabPropDisj.

    Variable b0 : branch.
    Variable P : branch -> Prop.
    Hypothesis P0 : P b0.
    Hypothesis P_eq  : forall x y, x eq y -> P x -> P y.
    Hypothesis P_exp : expansion_stable P. 

    Proposition tableau_prop_disj : forall t, tableau b0 t -> exists b, carrier t b /\ P b.
    Proof.
      apply (tableau_prop_disj_rec b0 (fun lb => exists b, (carrier lb) b /\ P b)).
      exists b0; auto.
      intros ? ? ? ( b & ? & ? ).
        exists y. split; auto.
        apply P_eq with (1 := H).
        destruct H0 as [ H0 | [] ].
        rewrite H0; auto.
      intros l1 l2 l3 ( b & H1 & H2); 
        exists b; split; auto.
        apply carrier_app, bin_union_inc_r2, carrier_app; pred_solve.
      intros ? ? ( b & H & ? ).
        apply carrier_app, bin_union_sg in H. 
        destruct H; [ left | right ]; exists b; tauto.
      intros ? lb ( ? & H & Hb' ) ?.
        destruct H as [ H | H ]; (contradiction H || trivial).
        rewrite <- H in Hb'. 
        apply (P_exp _ lb) in Hb'; auto. clear x H.
        destruct Hb' as ( x & y & H1 & H2 & H3 ).
        exists y. split; auto.
        apply P_eq with (1 := H2); auto.
    Qed.

  End TabPropDisj.

  Section ClosedTabs.

    Variable closed : branch -> Prop.
    Hypothesis closed_eq : forall x y, x eq y -> closed x -> closed y.

    Definition has_closed_tableau b0 := exists t, tableau b0 t /\ carrier t inc closed.

    Proposition has_closed_tableau_eq b0 b1 : b0 eq b1 -> has_closed_tableau b0 -> has_closed_tableau b1.
    Proof.
      intros H1 ( t & H2 & H3 ).
      exists t.
      split; auto.
      apply tableau_eq with (3 := H2); auto.
    Qed.

    Let closed_tableau_comp_rec b lb' lb : tableau b (lb++lb') -> carrier lb inc closed -> carrier lb' inc has_closed_tableau -> has_closed_tableau b.
    Proof.
      generalize lb; clear lb; induction lb'; intros lb H0 H1 H2. 
      rewrite <- app_nil_end in H0; exists lb; split; auto.
      apply carrier_cons_1 in H2; destruct H2 as [ [ ta [ H2 H3 ] ] H4 ].
      apply (IHlb' (lb++ta)); auto.
      rewrite -> app_ass; apply tableau_comp with (2 := H0); auto.
    Qed.

    Proposition closed_tableau_comp b lb : tableau b lb -> carrier lb inc has_closed_tableau -> has_closed_tableau b.
    Proof. intros; apply (closed_tableau_comp_rec b lb nil); auto. Qed.

    Proposition closed_tableau_expansion b lb : expansion b lb -> carrier lb inc has_closed_tableau -> has_closed_tableau b.
    Proof. 
      intros H.
      apply closed_tableau_comp. 
      cutrewrite (lb = nil++lb++nil).
      apply tableau_next with (b := b), tableau_start; auto.
      simpl; apply app_nil_end.
    Qed.

    Proposition closed_tab_comp_1 b0 b : has_closed_tableau b0 -> tableau b (b0::nil) -> has_closed_tableau b.
    Proof. intros ? H; apply closed_tableau_comp with (1 := H); auto. Qed.
   
    Proposition closed_tab_comp_2 b0 b1 b : has_closed_tableau b0 -> has_closed_tableau b1 -> tableau b (b0::b1::nil) -> has_closed_tableau b.
    Proof. intros ? ? H; apply closed_tableau_comp with (1 := H); auto. Qed.
 
    Hypothesis has_closed_tableau_oracle : forall b, has_closed_tableau b \/ ~ has_closed_tableau b. 

    Proposition has_no_closed_tableau_exp_stable : expansion_stable (fun b => ~ has_closed_tableau b).
    Proof.
      intros b lb He Hb.
      destruct (carrier_select_dec _ _ has_closed_tableau_oracle lb) as [ H | H ]; auto.
      contradiction Hb. 
      apply closed_tableau_expansion with (2 := H); auto.
      destruct H as ( x & H1 & H2 ).
      exists x; exists x; auto.
    Qed.

    Proposition has_no_closed_tableau_exp_1 b b0 : ~ has_closed_tableau b -> expansion b (b0::nil) -> ~ has_closed_tableau b0.
    Proof.
      intros H H0.
      destruct (has_no_closed_tableau_exp_stable _ _ H0 H) as ( x & y & H1 & H2 & H3 ).
      destruct H1 as [ H1 | [] ]; rewrite -> H1; auto.
      contradict H3; apply has_closed_tableau_eq with (2 := H3); auto.
    Qed.

    Proposition has_no_closed_tableau_exp_2 b b0 b1 : ~ has_closed_tableau b -> expansion b (b0::b1::nil) -> ~ has_closed_tableau b0 \/ ~ has_closed_tableau b1.
    Proof.
      intros H H0.
      destruct (has_no_closed_tableau_exp_stable _ _ H0 H) as ( x & y & H1 & H2 & H3 ).
      destruct H1 as [ H1 | [ H1 | [] ] ]; rewrite -> H1; auto; [ left | right ];
      contradict H3; apply has_closed_tableau_eq with (2 := H3); auto.
    Qed.

    Variable realizable : branch -> Prop.
    Hypothesis realizable_eq  : forall x y, x eq y -> realizable x -> realizable y.
    Hypothesis realizable_exp : expansion_stable realizable.
    Hypothesis real_clo_incomp : forall b, realizable b -> closed b -> False.

    Proposition realizable_has_no_closed_tableau : realizable inc (fun b => ~ has_closed_tableau b).
    Proof.
      intros b Hb ( t & H1 & H2 ).
      apply (tableau_prop_disj _ _ Hb realizable_eq realizable_exp) in H1.
      destruct H1 as ( b' & Hb' & Hr ).
      apply H2 in Hb'.
      apply (real_clo_incomp b'); trivial.
    Qed.

  End ClosedTabs.
 
End Tableau.

Implicit Arguments expansion_stable [ branch ].
Implicit Arguments tableau [ branch ].
Implicit Arguments has_closed_tableau [ branch ].

Hint Resolve listbranch_eq_refl listbranch_eq_sym.

